/* **************************************************
* pin connections(LCD->AVR):
* VCC->+3.3
* GND->GND
* CS->PB0
* RST->PB1
* A0->PB2
* SDA->MOSI
* SCK->SCK
* LED->VCC
************************************************** */
#define F_CPU 16000000
#include "tft144.h"
#include <stdio.h>
#include <avr/io.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <time.h>
#define NUM_NOTE		40

int frecventa_nota[NUM_NOTE] = {
	659.25511, 493.8833, 523.25113, 587.32954, 523.25113, 493.8833, 440.0, 440.0,
	523.25113, 659.25511, 587.32954, 523.25113, 493.8833, 523.25113, 587.32954,
	659.25511, 523.25113, 440.0, 440.0, 440.0, 493.8833, 523.25113, 587.32954,
	698.45646, 880.0, 783.99087, 698.45646, 659.25511, 523.25113, 659.25511,
	587.32954, 523.25113, 493.8833, 493.8833, 523.25113, 587.32954, 659.25511,
	523.25113, 440.0, 440.0
};

void uart_putchar(char c);
char uart_getchar();

FILE uart_output = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);
FILE uart_input = FDEV_SETUP_STREAM(NULL,uart_getchar, _FDEV_SETUP_READ);

void uart_putchar(char c) {
	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0 = c;
	loop_until_bit_is_set(UCSR0A, TXC0); /* Wait until transmission ready. */
}

char uart_getchar() {
	loop_until_bit_is_set(UCSR0A,RXC0);
	return UDR0;
}

u8 level = 0;
u8 levelcomplete = 0;;


uint8_t table1[8][8] = {{2,6,5,3,4,8,7,1},
						{4,8,1,7,5,3,6,2},
						{5,2,7,6,8,4,1,3},
						{3,1,4,8,6,7,2,5},
						{1,4,8,2,7,5,3,6},
						{6,7,3,5,1,2,8,4},
						{8,3,6,4,2,1,5,7},
						{7,5,2,1,3,6,4,8}};

uint8_t conf1[8][8] =  {{0,1,0,1,1,0,0,1},
						{1,0,0,0,1,1,0,1},
						{0,0,0,1,0,0,0,0},
						{0,1,0,0,0,1,0,1},
						{0,0,1,1,0,0,0,0},
						{0,1,0,0,0,0,1,1},
						{0,0,0,1,0,1,0,0},
						{0,1,1,1,0,1,0,1}};

uint8_t conf2[8][8] =  {{0,0,0,0,1,1,1,0},
						{1,0,0,0,0,0,1,1},
						{0,0,1,0,1,0,0,1},
						{0,0,0,0,1,1,0,1},
						{0,1,1,1,0,0,1,0},
						{0,0,0,1,0,1,0,0},
						{0,0,1,1,0,0,1,1},
						{1,0,1,0,0,0,1,0}};

uint8_t conf3[8][8] =  {{0,0,0,0,1,0,1,0},
						{1,0,0,0,0,0,0,1},
						{0,0,1,1,1,0,0,0},
						{1,1,1,0,0,0,1,1},
						{0,1,0,0,1,1,0,0},
						{1,1,0,0,0,1,0,0},
						{0,0,0,1,0,1,0,0},
						{1,0,0,1,0,0,1,0}};

uint8_t table2[8][8] = {{8,5,7,3,1,6,2,4},
						{6,1,4,2,3,5,7,8},
						{2,7,1,5,8,3,4,6},
						{4,8,3,6,7,2,1,5},
						{7,2,5,8,4,1,6,3},
						{1,3,6,4,5,7,8,2},
						{3,6,8,1,2,4,5,7},
						{5,4,2,7,6,8,3,1}};

uint8_t table3[8][8] = {{2,4,7,6,5,3,8,1},
						{1,8,5,3,6,7,4,2},
						{7,3,4,5,2,1,6,8},
						{8,1,6,2,7,5,3,4},
						{6,2,3,8,1,4,7,5},
						{4,5,1,7,3,8,2,6},
						{5,7,2,4,8,6,1,3},
						{3,6,8,1,4,2,5,7}};

uint8_t check[8][8] =  {{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0}};

uint8_t final[8][8] =  {{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0},
						{0,0,0,0,0,0,0,0}};

/* Genereaza o nota folosind speaker-ul. */
void speaker_reda(int nota) {
	int perioada_nota = 1000000l / frecventa_nota[nota];
	int i;

	/*
	 * Nu putem folosi variable ca argument pentru functia
	 * _delay_us(), asa ca folosim acest "workaround".
	 */
	PORTD |= (1 << PD4);
	for (i = 0; i < perioada_nota / 2; ++i)
		_delay_us(1);

	PORTD &= ~(1 << PD4);
	for (i = 0; i < perioada_nota / 2; ++i)
		_delay_us(1);
}

void play(int x){
	/* Setez pinul PD4 ca iesire. */
	DDRD |= (1 << PD4);
	int nota_curenta = -1;
	/* Opresc speaker-ul. */
	PORTD &= ~(1 << PD4);
	while(1){	

			if (nota_curenta < NUM_NOTE - 1)
			++nota_curenta;
			_delay_ms(1);
			if (nota_curenta != -1) {
				/* Redam nota. */
				for(int i = 0;i < x;i++)
				speaker_reda(nota_curenta);
				}
			if(nota_curenta == NUM_NOTE-1){
				nota_curenta=0;
				break;
			}
		}
}
void USART_Init() {
	UCSR0A=0x00;
	UCSR0B=0x18;
	UCSR0C=0x86;
	UBRR0H=0x00;
	UBRR0L=0x05;
}

void init() {
	PORTB = (1 << 4);
	DDRB = (1 << 7) | (1 << 5) | (1 << 4) | 1; // (MOSI,SCK,SS) should be output
	stdout = &uart_output;
	stdin = &uart_input;
	USART_Init(115200);
	printf("initializing system...\r\n");
	tft144_reset();
	tft144_init();
}

void drawTable(){
	tft144_draw_rectangle(0,0,127,127,WHITE);
	tft144_draw_line(15,0,15,127,WHITE);
	tft144_draw_line(31,0,31,127,WHITE);
	tft144_draw_line(47,0,47,127,WHITE);
	tft144_draw_line(63,0,63,127,WHITE);
	tft144_draw_line(79,0,79,127,WHITE);
	tft144_draw_line(95,0,95,127,WHITE);
	tft144_draw_line(111,0,111,127,WHITE);
	tft144_draw_line(127,0,127,127,WHITE);
	tft144_draw_line(0,15,127,15,WHITE);
	tft144_draw_line(0,31,127,31,WHITE);
	tft144_draw_line(0,47,127,47,WHITE);
	tft144_draw_line(0,63,127,63,WHITE);
	tft144_draw_line(0,79,127,79,WHITE);
	tft144_draw_line(0,95,127,95,WHITE);
	tft144_draw_line(0,111,127,111,WHITE);
	tft144_draw_line(0,127,127,127,WHITE);
}

void selectTable(int i, int j, int lev){
	char ch[2];
	u8 a=i*16+5,b=j*16+2;
	if(lev == 0){
		sprintf(ch,"%u",table1[i][j]);
		final[i][j]=table1[i][j];
	}
	else if(lev == 1){
		sprintf(ch,"%u",table2[i][j]);
		final[i][j]=table2[i][j];
	}
	else if(lev == 2){
		sprintf(ch,"%u",table3[i][j]);
		final[i][j]=table3[i][j];
	}
	tft144_putstring(ch,&a,&b,WHITE,BLACK,2);
	check[i][j]=1;
	//final[i][j]=table1[i][j];
}


void putNumbers(int lev){
	uint8_t i, j;
	if(lev == 0){
		for(i=0;i<8;i++)
		for(j=0;j<8;j++)
		if(conf1[i][j] == 1)
		selectTable(i,j,lev);
	}
	else if(lev == 1){
		for(i=0;i<8;i++)
		for(j=0;j<8;j++)
		if(conf2[i][j] == 1)
		selectTable(i,j,lev);
	}
	else if(lev == 2){
		for(i=0;i<8;i++)
		for(j=0;j<8;j++)
		if(conf3[i][j] == 1)
		selectTable(i,j,lev);
	}
}

void clearScreen(){
	tft144_clear_display(BLACK);
	uint8_t i, j;
	if(level == 0){
		for(i=0;i<8;i++)
		for(j=0;j<8;j++){
			if(conf1[i][j] == 1)
				final[i][j] = table1[i][j];
			else 
				final[i][j] = 0;
			check[i][j] = conf1[i][j];
		}
		
		
	}
	else if(level == 1){
		for(i=0;i<8;i++)
		for(j=0;j<8;j++){
			if(conf2[i][j] == 1)
				final[i][j] = table2[i][j];
			else 
				final[i][j] = 0;
			check[i][j] = conf2[i][j];
		}
	}
	else if(level == 2){
		for(i=0;i<8;i++)
		for(j=0;j<8;j++){
			if(conf3[i][j] == 1)
				final[i][j] = table3[i][j];
			else 
				final[i][j] = 0;
			check[i][j] = conf3[i][j];
		}
	}
}

void autoSOlve(){
		uint8_t i, j;
		if(level == 0){
			for(i=0;i<8;i++)
			for(j=0;j<8;j++)
			if(conf1[i][j] == 0)
			selectTable(i,j,level);
		}
		else if(level == 1){
			for(i=0;i<8;i++)
			for(j=0;j<8;j++)
			if(conf2[i][j] == 0)
			selectTable(i,j,level);
		}
		else if(level == 2){
			for(i=0;i<8;i++)
			for(j=0;j<8;j++)
			if(conf3[i][j] == 0)
			selectTable(i,j,level);
		}
}

void checkSolution(){
	u8 a = 1,i,j;
	for(i=0;i<8;i++)
	for(j=0;j<8;j++){
		if(level == 0){
			if(final[i][j] != table1[i][j])
				a = 0;
		}
		if(level == 1){
			if(final[i][j] != table2[i][j])
				a = 0;
		}
		if(level == 2){
			if(final[i][j] != table3[i][j])
				a = 0;
		}
	}
	if(a == 1)
		levelcomplete = 1;
}

int main() {
	u8 a,b,c,d, e,f;
	DDRA=0xF0;
	PORTA=0x0F;
	init();
	for(a=0;a<128;a++){
		for(b=0;b<128;b++)
		tft144_draw_dot(a,b,RGB565(a*a,b*b,a*b));
	}
	_delay_ms(1000);
	tft144_clear_display(RED);
	u8 x=40,y=40;
	tft144_putstring("SUDOKU", &x, &y, BLACK, RED,2);
	_delay_ms(5000);
	play(10);
	tft144_clear_display(BLACK);

	a=b=0;
	c=d=15;
	e=f=0;
	
	while(1) {
		drawTable();
		putNumbers(level);
		tft144_draw_rectangle(a,b,c,d,RED);	
		_delay_ms(100);

		e=a+5;f=b+2;
		
		PORTA =0b01111111;
		/* Tasta * de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 3))
		{
			//tft144_putstring("*",&e,&f,GREEN,BLACK,2);
			while(!bit_is_set(PINA, 3));
		}

		/* Tasta C de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA,2))
		{
			if(b>0){
				b-=16;
				d-=16;
			}
			//tft144_putstring("C",&e,&f,GREEN,BLACK,2);
			while(!bit_is_set(PINA, 2));
			
		}

		/* Tasta B de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 1))
		{
			if(d < 127){
				b+=16;
				d+=16;
			}
			//tft144_putstring("B",&e,&f,GREEN,BLACK,2);
			while(!bit_is_set(PINA, 1));
		}

		/* Tasta A de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 0))
		{
			if(a>0){
				a-=16;
				c-=16;
			}
			//tft144_putstring("A",&e,&f,GREEN,BLACK,2);
			while(!bit_is_set(PINA, 0));
		}
		
		/* A treia coloana a keypad-ului 4x4. */
		PORTA =0b10111111;
		
		/* Tasta D de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 3))
		{
			if(c<127){
				a+=16;
				c+=16;
			}
			//tft144_putstring("D",&e,&f,GREEN,BLACK,2);
			while(!bit_is_set(PINA, 3));
		}
		
		/* Tasta 9 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 2))
		{
			//tft144_putstring("9",&e,&f,GREEN,BLACK,2);
			while(!bit_is_set(PINA, 2));
		}

		/* Tasta 6 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 1))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("6", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=6;
			}
			while(!bit_is_set(PINA, 1));
		}

		/* Tasta 3 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 0))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("3", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=3;
			}
			while(!bit_is_set(PINA, 0));
		}
		
		/* A doua coloana a keypad-ului 4x4. */
		PORTA =0b11011111;
		
		/* Tasta # de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 3))
		{
			//tft144_putstring("#",&e,&f,GREEN,BLACK,2);
			autoSOlve();
			while(!bit_is_set(PINA, 3));
		}
		
		/* Tasta 8 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 2))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("8", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=8;
			}
			while(!bit_is_set(PINA, 2));
		}
		
		/* Tasta 5 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 1))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("5", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=5;
			}
			while(!bit_is_set(PINA, 1));
		}
		
		/* Tasta 2 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 0))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("2", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=2;
			}
			while(!bit_is_set(PINA, 0));
		}
		
		/* Prima coloana a keypad-ului 4x4. */
		PORTA =0b11101111;

		/* Tasta 0 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 3))
		{
			//tft144_putstring("0",&e,&f,GREEN,BLACK,2);
			clearScreen();
			while(!bit_is_set(PINA, 3));
		}


		/* Tasta 7 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 2))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("7", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=7;
			}
			while(!bit_is_set(PINA, 2));
		}

		/* Tasta 4 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 1))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("4", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=4;
			}
			while(!bit_is_set(PINA, 1));
		}

		/* Tasta 1 de pe keypad-ul 4x4. */
		if (!bit_is_set(PINA, 0))
		{
			if(check[a/16][b/16] !=1){
				check[a/16][b/16] = 1;
				tft144_putstring("1", &e,&f,GREEN,BLACK,2);
				final[a/16][b/16]=1;
			}
			while(!bit_is_set(PINA, 0));
		}

		checkSolution();
		if(levelcomplete == 1) {
			play(60);
			_delay_ms(1000);
			level++;
			clearScreen();
			levelcomplete = 0;
		}

		if(level > 2)
			break;
		
	}

	tft144_clear_display(RED);
	while(1){
		u8 x=15,y=40;
		tft144_putstring("FELICITARI!!!", &x, &y, BLACK, RED,2);
	}
	
	return 0;
}

